/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "outlook/utils.h"



static BOOL dbg = false;


/*
* The method return current system time from VariantTime format 
*
* @param tempDate : the variable to write date in
* @param date     : the date in variant time format
*/
void doubleToSystemTime(wchar_t* tempDate, DATE date) {
    doubleToSystemTimeHourZero(tempDate, date, FALSE);
}

/*
* Is like doubleToSystemTime but set hour to 0. It used directly for anniversary and birthday.
* if not the time represented by date is the same in which the date is written 
* on the device
*/
void doubleToSystemTimeHourZero(wchar_t* tempDate, DATE date, BOOL isHourToSetToZero) {
        
    SYSTEMTIME t;
    BOOL isUTC = TRUE;
    VariantTimeToSystemTime(date, &t); 
    
    if (isHourToSetToZero) {
        t.wHour = 0;
        t.wMinute = 0;
        t.wSecond = 0;
    }
    

    // conversion from actual time to UTC    
    //wchar_t propertyValue[10];
    //getClientConfigurationInternal (NULL, PROPERTY_SOURCEPROP_SPDM_SYSTEMTIME_TO_LOCALTIME, propertyValue, FALSE); 

    //if (wcscmp(propertyValue, TEXT("1")) == 0) {
    //    isUTC = TRUE;
    //}
    
    if (isUTC) {
        localTimeToUTC(t);
    }

    wchar_t month [10];
    wchar_t day   [10];
    wchar_t hour  [10];
    wchar_t minute[10];
    wchar_t second[10];
    
    // month
    if (t.wMonth < 10)
        wsprintf(month, TEXT("0%i"), t.wMonth);
    else
        wsprintf(month, TEXT("%i"), t.wMonth);
    
    // day
    if (t.wDay  < 10)
        wsprintf(day, TEXT("0%i"), t.wDay);
    else
        wsprintf(day, TEXT("%i"), t.wDay);
    
    // hour
    if (t.wHour  < 10)
        wsprintf(hour, TEXT("0%i"), t.wHour);
    else
        wsprintf(hour, TEXT("%i"), t.wHour);

    // minute
    if (t.wMinute < 10)
        wsprintf(minute, TEXT("0%i"), t.wMinute);
    else
        wsprintf(minute, TEXT("%i"), t.wMinute);

    // second
    if (t.wSecond < 10)
        wsprintf(second, TEXT("0%i"), t.wSecond);
    else
        wsprintf(second, TEXT("%i"), t.wSecond);
    
    if (isUTC) {
        wsprintf(tempDate, TEXT("%i%s%sT%s%s%sZ"), t.wYear, month, day, hour, minute, second);
    } else {
        wsprintf(tempDate, TEXT("%i%s%sT%s%s%s"), t.wYear, month, day, hour, minute, second);
    }   

}



/*
 * Converts a date in double format into "yyyyMMdd" format. 
 * All-day-event items need this date format.
*/
void doubleToSystemTimeOnlyDate(wchar_t* tempDate, DATE date) {
        
    SYSTEMTIME t;
    VariantTimeToSystemTime(date, &t); 


    wchar_t month [10];
    wchar_t day   [10];

    
    // month
    if (t.wMonth < 10)
        wsprintf(month, TEXT("0%i"), t.wMonth);
    else
        wsprintf(month, TEXT("%i"), t.wMonth);
    
    // day
    if (t.wDay  < 10)
        wsprintf(day, TEXT("0%i"), t.wDay);
    else
        wsprintf(day, TEXT("%i"), t.wDay);
    

    wsprintf(tempDate, TEXT("%i%s%s"), t.wYear, month, day);
}



/*
* Is like doubleToSystemTime but set hour to 0. It used directly for anniversary and birthday.
* if not the time represented by date is the same in which the date is written 
* on the device
*/
void doubleToSystemTimeBirthday(wchar_t* tempDate, DATE date) {
        
    SYSTEMTIME t;
    
    VariantTimeToSystemTime(date, &t); 
    
    t.wHour = 0;
    t.wMinute = 0;
    t.wSecond = 0;

    wchar_t month [10];
    wchar_t day   [10];
    
    // month
    if (t.wMonth < 10)
        wsprintf(month, TEXT("0%i"), t.wMonth);
    else
        wsprintf(month, TEXT("%i"), t.wMonth);
    
    // day
    if (t.wDay  < 10)
        wsprintf(day, TEXT("0%i"), t.wDay);
    else
        wsprintf(day, TEXT("%i"), t.wDay);
    
    
    wsprintf(tempDate, TEXT("%i-%s-%s"), t.wYear, month, day);
    

}



/*
* The method transform a date in format "YYYYMMDDThhmmss" or TEXT("YYYYMMDDThhmmssZ") with hour 0-24 into VariantTime format.
*
* @param dataString : the string in "YYYYMMDDThhmmss" TEXT("YYYYMMDDThhmmssZ") format
* @param date       : the returned value into VariantTime format
* @action           : the parameter to set manually hour and minute. If NULL current value are used
*/
void systemTimeToDouble(wchar_t* dataString, DATE* date, wchar_t* action) {

    wchar_t* token  = NULL;
    wchar_t* dummy  = NULL;
    wchar_t* Ztoken = NULL;
    BOOL isUTC      = FALSE;
    wchar_t internalString[DIM_LOCAL_TEMP];
    wsprintf(internalString, dataString);
    wchar_t tmp[50];
    
    int ret = 0;
    
    SYSTEMTIME t;
    TIME_ZONE_INFORMATION tmz;        
    
    GetTimeZoneInformation(&tmz);
    
    //  TEXT("20041006T203457") or TEXT("20041006T203457Z")
    token = wcstok(internalString, TEXT("T"));
    Ztoken = wcschr(dataString, 'Z');
    if (Ztoken != NULL) {
        isUTC = TRUE;
    }
    
    if (token != NULL) {
            
        //yyyy
        wcsncpy(tmp, dataString, 4);
        tmp[4] = 0;
        swscanf(tmp, TEXT("%d"),&t.wYear);
        //MM
        wcsncpy(tmp, dataString + 4, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wMonth);
        //dd
         wcsncpy(tmp, dataString + 6, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wDay);

        if (action != NULL && wcscmp(action, TEXT("0000")) == 0) {
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
            
        } else if (action != NULL && wcscmp(action, TEXT("2359")) == 0) {
            t.wHour = 23;
            t.wMinute = 59;
            t.wSecond = 59;
        } else {
            if (wcslen(dataString) > 9) {
                //HH
                wcsncpy(tmp, dataString + 9, 2);
                tmp[2] = 0;
                swscanf(tmp, TEXT("%d"),&t.wHour);
                                
                //mm
                 wcsncpy(tmp, dataString + 11, 2);
                tmp[2] = 0;
                swscanf(tmp, TEXT("%d"),&t.wMinute);
                //ss
                 wcsncpy(tmp, dataString + 13, 2);
                tmp[2] = 0;
                swscanf(tmp, TEXT("%d"),&t.wSecond);
                

            } else {
                t.wHour = 0;
                t.wMinute = 0;
                t.wSecond = 0;
            }

        }
        // set to default
        t.wMilliseconds = 0;                    
        t.wDayOfWeek = 0;                       
                
        
        if(isUTC) {
            UTCToLocalTime(t);
        }
        ret = SystemTimeToVariantTime (&t, date);
       
            
    } else {
        
        (*date) = 949998.000000;
        
    }

}




/*
* The method transform a string date in format "yyyy-MM-dd" to DATE
*
* @param dataString : the string in yyyy-MM-dd format
* @param date       : the returned value into VariantTime format
* @action           : the parameter to set manually hour and minute. If NULL current value are used
*/
void systemTimeToDoubleBirthday(wchar_t* dataString, DATE* date) {
    systemTimeToDoubleBirthFormat(dataString, date, TEXT("START"));
}

/*
* It is used to transform yyyy-MM-dd format date into DATE. type parameter is to set
* the hours and minutes. 
* Parameter END to set hour 23:59
* Parameter START to set hour 00:00
* Set the hour to 00:00 is used for birthday, anniversary and startDate of appoitment
* when it is all day event. Set hour 23:59 for endDate when an appointment it is allDayEvent
*/
void systemTimeToDoubleBirthFormat(wchar_t* dataString, DATE* date, wchar_t* type) {

    wchar_t* token  = NULL;
    wchar_t* dummy  = NULL;
    wchar_t* Ztoken = NULL;
    
    wchar_t internalString[DIM_LOCAL_TEMP];
    wsprintf(internalString, dataString);
    wchar_t tmp[50];
    
    int ret = 0;
    
    SYSTEMTIME t;
    TIME_ZONE_INFORMATION tmz;        
    
    GetTimeZoneInformation(&tmz);
    
    Ztoken = wcschr(dataString, '-');
    if (Ztoken == NULL) {
        systemTimeToDouble(dataString, date, NULL);
        return;
    }
    
    //  TEXT("yyyy-MM-dd") 
    token = wcstok(internalString, TEXT("-"));

    if (token != NULL) {
            
        //yyyy
        wcsncpy(tmp, dataString, 4);
        tmp[4] = 0;
        swscanf(tmp, TEXT("%d"),&t.wYear);
        //MM
        wcsncpy(tmp, dataString + 5, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wMonth);
        //dd
         wcsncpy(tmp, dataString + 8, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wDay);
        
        if (wcscmp(type, TEXT("END")) == 0) {
        
            t.wHour = 23;
            t.wMinute = 59;
            t.wSecond = 0;
        } else {
        
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
        } 
                   
        // set to default
        t.wMilliseconds = 0;                    
        t.wDayOfWeek = 0;                       
        
        ret = SystemTimeToVariantTime (&t, date);
       
            
    } else {
        
        (*date) = 949998.000000;
        
    }
}




/////////////////////////////////// For recurrence ////////////////////////////////////

/*
 * Converts recurrence data from Rrule (Ical) format to SIF.
 * 
 * @param recurrenceString  : output string in SIF format (initiated inside the function)
 * @param rrule             : input data in Rrule format
 * @param startDate			: the start date of current recurrence (double format)
 * @return					: 0 (false) if error occurred
 */
bool convertRruleToSIF(string &recurrenceString, wchar_t* rrule, DATE startDate) {
	
	bool ret = true;
	wchar_t tempbuf[DIM_LOCAL_TEMP];
	wchar_t* str = new wchar_t[100];
	wcscpy(str, rrule);
	wchar_t seps[] = TEXT(" ");
	wchar_t* token;

	recurrenceString = TEXT("");

	token = wcstok( str, seps );

	//first will be D|W|MP|MD|YM|YD<interval>
	long recType = -1;
	long interval;
	long occurences;
	//DATE endDate;
	long dayOfMonth = -1;
	long weekOfMonth = -1;
	long monthOfYear = -1;

	if(token[0] == TEXT('D')) {
		recType = olRecursDaily;
		token++;
	}
	else if(token[0] == TEXT('W')) {
		recType = olRecursWeekly;
		token++;
		
	}
	else if(token[0] == TEXT('M') && token[1] == TEXT('D')) {
		recType = olRecursMonthly;
		token = token + 2;
	}
	else if(token[0] == TEXT('M') && token[1] == TEXT('P')) {
		recType = olRecursMonthNth;
		token = token + 2;
	}
	//not supported
	else if(token[0] == TEXT('Y') && token[1] == TEXT('D'))
		recType = -1;
	else if(token[0] == TEXT('Y') && token[1] == TEXT('M')) {		
		recType = olRecursYearly; //!!!
		token = token + 2;
	}

	interval = _wtol(token);
	if(!interval || recType == -1) {
		if(str)
			delete [] str;
		return false;
	}

	wchar_t* days = new wchar_t[20];
	wcscpy(days, TEXT(""));

	wchar_t* mOfYear = new wchar_t[20];
	wcscpy(mOfYear, TEXT(""));

	while( token != NULL ) {
		token = wcstok( NULL, seps );
		if(token) {

			if(recType == olRecursDaily) {
				//this toke should be even number of occurrences or the end date of the pattern				
				if(wcschr(token, TEXT('#'))) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					token++;
					occurences = _wtol(token);
					if(occurences == 0){
						wsprintf(tempbuf, TEXT("%i"), TRUE);
						recurrenceString  = recurrenceString + TEXT("<NoEndDate>") + tempbuf + TEXT("</NoEndDate>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), occurences);
						recurrenceString  = recurrenceString + TEXT("<Occurrences>") + tempbuf + TEXT("</Occurrences>");
					}
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					
				}
				else if(token[8] == TEXT('T')) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					recurrenceString  = recurrenceString + TEXT("<PatternEndDate>") + token + TEXT("</PatternEndDate>");
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
				}
			}

			else if(recType == olRecursWeekly) {				
				if(wcschr(token, TEXT('#'))) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					token++;
					occurences = _wtol(token);
					if(occurences == 0){
						wsprintf(tempbuf, TEXT("%i"), TRUE);
						recurrenceString  = recurrenceString + TEXT("<NoEndDate>") + tempbuf + TEXT("</NoEndDate>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), occurences);
						recurrenceString  = recurrenceString + TEXT("<Occurrences>") + tempbuf + TEXT("</Occurrences>");
					}
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");

					if(wcscmp(days, TEXT(""))) {
						wsprintf(tempbuf, TEXT("%i"), calculateDayOfWeek(days));
						recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), getWeekDay(startDate));
						recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
					}
				}
				else if(token[8] == TEXT('T')) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					recurrenceString  = recurrenceString + TEXT("<PatternEndDate>") + token + TEXT("</PatternEndDate>");
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					if(wcscmp(days, TEXT(""))) {
						wsprintf(tempbuf, TEXT("%i"), calculateDayOfWeek(days));
						recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), getWeekDay(startDate));
						recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
					}
				}
				else if(isDayOfWeek(token)) {
 					wcscat(days, token);
					wcscat(days, TEXT(" "));
				}
			}

			else if(recType == olRecursMonthly) {
				//VCalendar supports as Monthly by day rules: MD<interval> listofdates|listofdates(from end of month) <end tag>
				//On outlook recurrence pattern only MD<interval> day <end tag> can be represented
				if(wcschr(token, TEXT('#'))) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					token++;
					occurences = _wtol(token);
					if(occurences == 0) {
						wsprintf(tempbuf, TEXT("%i"), TRUE);
						recurrenceString  = recurrenceString + TEXT("<NoEndDate>") + tempbuf + TEXT("</NoEndDate>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), occurences);
						recurrenceString  = recurrenceString + TEXT("<Occurrences>") + tempbuf + TEXT("</Occurrences>");
					}
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), dayOfMonth);
					recurrenceString  = recurrenceString + TEXT("<DayOfMonth>") + tempbuf + TEXT("</DayOfMonth>");
				}
				else if(token[8] == TEXT('T')) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					recurrenceString  = recurrenceString + TEXT("<PatternEndDate>") + token + TEXT("</PatternEndDate>");
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), dayOfMonth);
					recurrenceString  = recurrenceString + TEXT("<DayOfMonth>") + tempbuf + TEXT("</DayOfMonth>");
				}
				else {
					if(dayOfMonth != -1) {
						if(str)
							delete [] str;
						return false;
					}
					dayOfMonth = _wtol(token);
					if(dayOfMonth == 0) {
						if(str)
							delete [] str;
						return false;
					}
				}
			}

			else if(recType == olRecursMonthNth) {
				if(wcschr(token, TEXT('#'))) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					token++;
					occurences = _wtol(token);
					if(occurences == 0) {
						wsprintf(tempbuf, TEXT("%i"), TRUE);
						recurrenceString  = recurrenceString + TEXT("<NoEndDate>") + tempbuf + TEXT("</NoEndDate>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), occurences);
						recurrenceString  = recurrenceString + TEXT("<Occurrences>") + tempbuf + TEXT("</Occurrences>");
					}
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), weekOfMonth);
					recurrenceString  = recurrenceString + TEXT("<Instance>") + tempbuf + TEXT("</Instance>");
					wsprintf(tempbuf, TEXT("%i"), calculateDayOfWeek(days));
					recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
				}
				else if(token[8] == TEXT('T')) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					recurrenceString  = recurrenceString + TEXT("<PatternEndDate>") + token + TEXT("</PatternEndDate>");
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), weekOfMonth);
					recurrenceString  = recurrenceString + TEXT("<Instance>") + tempbuf + TEXT("</Instance>");
					wsprintf(tempbuf, TEXT("%i"), calculateDayOfWeek(days));
					recurrenceString  = recurrenceString + TEXT("<DayOfWeekMask>") + tempbuf + TEXT("</DayOfWeekMask>");
				}
				else if(isDayOfWeek(token)) {
 					wcscat(days, token);
					wcscat(days, TEXT(" "));
				}
				else { 
					if(token[1] != TEXT('+') && token[1] != TEXT('-')) {
						if(str)
							delete [] str;
						return false;
					}
					wchar_t sWeek[] = TEXT("\0\0");
					sWeek[0] = token[0];
					weekOfMonth = _wtol(sWeek);
					if(token[1] == TEXT('-'))
						weekOfMonth = 5 - weekOfMonth;
					if(weekOfMonth > 5 || weekOfMonth < 1) {
						if(str)
							delete [] str;
						return false;
					}
				}
			}

			else if(recType == olRecursYearly) {
				//expected sequence will be YM1 month <end tag>
				if(wcschr(token, TEXT('#'))) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					token++;
					occurences = _wtol(token);
					if(occurences == 0) {
						wsprintf(tempbuf, TEXT("%i"), TRUE);
						recurrenceString  = recurrenceString + TEXT("<NoEndDate>") + tempbuf + TEXT("</NoEndDate>");
					}
					else {
						wsprintf(tempbuf, TEXT("%i"), occurences);
						recurrenceString  = recurrenceString + TEXT("<Occurrences>") + tempbuf + TEXT("</Occurrences>");
					}
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), monthOfYear);
					recurrenceString  = recurrenceString + TEXT("<MonthOfYear>") + tempbuf + TEXT("</MonthOfYear>");
					wsprintf(tempbuf, TEXT("%i"), dayOfMonth);
					recurrenceString  = recurrenceString + TEXT("<DayOfMonth>") + tempbuf + TEXT("</DayOfMonth>");
				}
				else if(token[8] == TEXT('T')) {
					wsprintf(tempbuf, TEXT("%i"), recType);
					recurrenceString  = recurrenceString + TEXT("<RecurrenceType>") + tempbuf + TEXT("</RecurrenceType>");
					recurrenceString  = recurrenceString + TEXT("<PatternEndDate>") + token + TEXT("</PatternEndDate>");
					wsprintf(tempbuf, TEXT("%i"), interval);
					recurrenceString  = recurrenceString + TEXT("<Interval>") + tempbuf + TEXT("</Interval>");
					wsprintf(tempbuf, TEXT("%i"), monthOfYear);
					recurrenceString  = recurrenceString + TEXT("<MonthOfYear>") + tempbuf + TEXT("</MonthOfYear>");
					wsprintf(tempbuf, TEXT("%i"), dayOfMonth);
					recurrenceString  = recurrenceString + TEXT("<DayOfMonth>") + tempbuf + TEXT("</DayOfMonth>");
				}
				else {
					if(wcscmp(mOfYear, TEXT(""))) {
						if(str)
							delete [] str;
						return false;
					}
 					wcscat(mOfYear, token);
					monthOfYear = _wtol(mOfYear);
					SYSTEMTIME st;
					VariantTimeToSystemTime(startDate, &st);
					dayOfMonth = st.wDay;
				}
			}
		}		
	}

	return ret;
}

long calculateDayOfWeek(wchar_t* in) {

	long ret = 0;
	
	wchar_t* index;
	index = NULL;
	index = wcsstr(in, TEXT("SU"));
	if(index)
		ret += olSunday;

	index = NULL;
	index = wcsstr(in, TEXT("MO"));
	if(index)
		ret += olMonday;

	index = NULL;
	index = wcsstr(in, TEXT("TU"));
	if(index)
		ret += olTuesday;

	index = NULL;
	index = wcsstr(in, TEXT("WE"));
	if(index)
		ret += olWednesday;

	index = NULL;
	index = wcsstr(in, TEXT("TH"));
	if(index)
		ret += olThursday;

	index = NULL;
	index = wcsstr(in, TEXT("FR"));
	if(index)
		ret += olFriday;

	index = NULL;
	index = wcsstr(in, TEXT("SA"));
	if(index)
		ret += olSaturday;

	return ret; 
}


bool isDayOfWeek(wchar_t* string) {
	
	bool ret = false;
	wchar_t* weekDay[] = {TEXT("SU"), TEXT("MO"), TEXT("TU"), TEXT("WE"), TEXT("TH"), TEXT("FR"), TEXT("SA")};

	for(int i = 0; i < 7 ; i++) 
		if(!wcscmp(string, weekDay[i]))
			return true;

	return ret;
}

long getWeekDay(DATE  dt) {
	long ret = 0;

	SYSTEMTIME st;
	VariantTimeToSystemTime(dt, &st);

    ret = (long)(st.wDayOfWeek)*(st.wDayOfWeek);

	return ret;
}


/*
 * Converts recurrence data from SIF format to Rrule (Ical).
 * Returns a new allocated wchar_t*, must be freed by the caller.
 *
 */
wchar_t* convertSIFToRrule(wchar_t* ptrData) {
	
	wchar_t* datev = new wchar_t[30];
	wchar_t* element;

	wchar_t* recurrence = new wchar_t[100];
	wcscpy(recurrence, TEXT(""));

	
	long recType;
	element = getElementContent(ptrData, TEXT ("RecurrenceType"), NULL);
	recType = _wtoi(element);
	delete [] element; element = NULL;

	DATE patEndDate;
	element = getElementContent(ptrData, TEXT ("PatternEndDate"), NULL);
    systemTimeToDouble(element, &patEndDate, NULL);
	delete [] element; element = NULL;

	DATE patStartDate;
	element = getElementContent(ptrData, TEXT ("PatternStartDate"), NULL);
    systemTimeToDouble(element, &patStartDate, NULL);
	delete [] element; element = NULL;
	
	BOOL noEnd;
	element = getElementContent(ptrData, TEXT ("NoEndDate"), NULL);
	noEnd = _wtoi(element);
	delete [] element; element = NULL;

	long interval;
	element = getElementContent(ptrData, TEXT ("Interval"), NULL);
	interval = _wtoi(element);
	delete [] element; element = NULL;

	long occurences;
	element = getElementContent(ptrData, TEXT ("Occurrences"), NULL);
	occurences = _wtoi(element);
	delete [] element; element = NULL;
	
	long dayofweek;
	element = getElementContent(ptrData, TEXT ("DayOfWeekMask"), NULL);
	dayofweek = _wtoi(element);
	delete [] element; element = NULL;

	long dayofmonth;
	element = getElementContent(ptrData, TEXT ("DayOfMonth"), NULL);
	dayofmonth = _wtoi(element);
	delete [] element; element = NULL;

	//week of month
	long instance;
	element = getElementContent(ptrData, TEXT ("Instance"), NULL);
	instance = _wtoi(element);
	delete [] element; element = NULL;

	long monthofyear;
	element = getElementContent(ptrData, TEXT ("MonthOfYear"), NULL);
	monthofyear = _wtoi(element);
	delete [] element; element = NULL;


	switch(recType) {
			case olRecursDaily:
				if(noEnd == VARIANT_TRUE)
					occurences = 0;  //occurences are -1 in this case but we use 0 to indicate event that does not end
				if(interval >0)
						wsprintf(recurrence, TEXT("D%ld #%ld"), interval, occurences);
				else if (dayofweek > 0) {
					wchar_t* days = extractDayOfWeek(dayofweek);
					if (days) {
						doubleToSystemTimeHourZero(datev, patEndDate, FALSE);
						wsprintf(recurrence, TEXT("W1 %s %s #%ld"), days, datev, occurences);
						delete [] days;
					}
				}
				break;
			case olRecursWeekly:
				if (dayofweek > 0) {
					wchar_t* days = extractDayOfWeek(dayofweek);
					if (days) {
						if(noEnd == VARIANT_TRUE)
							wsprintf(recurrence, TEXT("W%ld %s #0"), interval, days);
						else {
							doubleToSystemTimeHourZero(datev, patEndDate, FALSE);
							wsprintf(recurrence, TEXT("W%ld %s%s #%ld"), interval, days, datev, occurences);
							delete [] days;
						}
					}
				}
				break;
			case olRecursMonthly:
				if(dayofmonth > 0) 
					if(noEnd == VARIANT_TRUE)
							wsprintf(recurrence, TEXT("MD%ld %ld #0"), interval, dayofmonth);
						else {
							doubleToSystemTimeHourZero(datev, patEndDate, FALSE);
                            wsprintf(recurrence, TEXT("MD%ld %ld %s #%ld"), interval, dayofmonth, datev, occurences);
						}
				break;
			case olRecursMonthNth:
				if(instance > 0 && dayofweek >0) {
					wchar_t* days = extractDayOfWeek(dayofweek);
					if(days)
						if(noEnd == VARIANT_TRUE)
								wsprintf(recurrence, TEXT("MP%ld %ld+ %s #0"), interval, instance, days);
							else {
								doubleToSystemTimeHourZero(datev, patEndDate, FALSE);
								wsprintf(recurrence, TEXT("MP%ld %ld+ %s%s #%ld"), interval, instance, days, datev, occurences);
							}
					if(days)
						delete [] days;
				}
				break;
			case olRecursYearly:
				if(dayofmonth > 0 && monthofyear > 0) 
					if(noEnd == VARIANT_TRUE) 
						wsprintf(recurrence, TEXT("YM1 %ld #0"), monthofyear);
					else
						wsprintf(recurrence, TEXT("YM1 %ld #%ld"), monthofyear, occurences);
				break;
			case olRecursYearNth:
				if(dayofweek > 0 && instance >0) {
					wchar_t* days = extractDayOfWeek(dayofweek);
					if(days)
						if(noEnd == VARIANT_TRUE)
							wsprintf(recurrence, TEXT("MP12 %ld+ %s #0"), instance, days);
							else {
								doubleToSystemTimeHourZero(datev, patEndDate, FALSE);
								wsprintf(recurrence, TEXT("MP12 %ld+ %s%s #%ld"), instance, days, datev, occurences);
							}
					if(days)
						delete [] days;
				}
				break;	
		}

		delete [] datev; datev = NULL;	

	return recurrence;
}



wchar_t* extractDayOfWeek(long l) {
	/*
		enum OlDaysOfWeek {
		  olSunday    = 1,
		  olMonday    = 2,
		  olTuesday   = 4,
		  olWednesday = 8,
		  olThursday  = 16,
		  olFriday    = 32,
		  olSaturday  = 64,
		};

		*/
	if (l<0 || l>128)
		return NULL;

	//SU MO TU WE TH FR SA

	wchar_t* ret = new wchar_t[21];
	wcscpy(ret, TEXT(""));
	
	if(l & olSunday)
		wcscat(ret, TEXT("SU "));
	if(l & olMonday)
		wcscat(ret, TEXT("MO "));
	if(l & olTuesday)
		wcscat(ret, TEXT("TU "));
	if(l & olWednesday)
		wcscat(ret, TEXT("WE "));
	if(l & olThursday)
		wcscat(ret, TEXT("TH "));
	if(l & olFriday)
		wcscat(ret, TEXT("FR "));
	if(l & olSaturday)
		wcscat(ret, TEXT("SA "));

	return ret;
	
}
////////////////////////////////////////////////////////////////////




/*
* return date and time of a date and if it is in utc or not
*/
BOOL getDateAndTimeAsIs(wchar_t* dataString, DATE* date) {

    wchar_t* token  = NULL;
    wchar_t* dummy  = NULL;
    wchar_t* Ztoken = NULL;
    BOOL isUTC      = FALSE;
    wchar_t internalString[DIM_LOCAL_TEMP];
    wsprintf(internalString, dataString);
    wchar_t tmp[50];
    
    int ret = 0;
    
    SYSTEMTIME t;
    TIME_ZONE_INFORMATION tmz;        
    
    GetTimeZoneInformation(&tmz);
    
    //  TEXT("20041006T203457") or TEXT("20041006T203457Z")
    token = wcstok(internalString, TEXT("T"));
    Ztoken = wcschr(dataString, 'Z');
    if (Ztoken != NULL) {
        isUTC = TRUE;
    }
    
    if (token != NULL) {
            
        //yyyy
        wcsncpy(tmp, dataString, 4);
        tmp[4] = 0;
        swscanf(tmp, TEXT("%d"),&t.wYear);
        //MM
        wcsncpy(tmp, dataString + 4, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wMonth);
        //dd
         wcsncpy(tmp, dataString + 6, 2);
        tmp[2] = 0;
        swscanf(tmp, TEXT("%d"),&t.wDay);
        
        if (wcslen(dataString) > 9) {
            //HH
            wcsncpy(tmp, dataString + 9, 2);
            tmp[2] = 0;
            swscanf(tmp, TEXT("%d"),&t.wHour);
                            
            //mm
             wcsncpy(tmp, dataString + 11, 2);
            tmp[2] = 0;
            swscanf(tmp, TEXT("%d"),&t.wMinute);
            //ss
             wcsncpy(tmp, dataString + 13, 2);
            tmp[2] = 0;
            swscanf(tmp, TEXT("%d"),&t.wSecond);
            
        } else {
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
        }

        
        // set to default
        t.wMilliseconds = 0;                    
        t.wDayOfWeek = 0;                       
                                
        ret = SystemTimeToVariantTime (&t, date);
       
            
    } else {
        
        (*date) = 949998.000000;
        
    }
    return isUTC;

}



void localTimeToUTC(SYSTEMTIME &sysTime)
{
     FILETIME LocalFileTime, SystemFileTime;
     SystemTimeToFileTime(&sysTime, &LocalFileTime);
     LocalFileTimeToFileTime(&LocalFileTime, &SystemFileTime);
     FileTimeToSystemTime(&SystemFileTime, &sysTime);
}



void UTCToLocalTime(SYSTEMTIME &sysTime)
{
     FILETIME LocalFileTime, SystemFileTime;
     SystemTimeToFileTime(&sysTime, &SystemFileTime);
     FileTimeToLocalFileTime(&SystemFileTime, &LocalFileTime);
     FileTimeToSystemTime(&LocalFileTime, &sysTime);
}








void encodeSpecialChar(string &s) {
        
    int position = 0, previousPosition = 0;
       
    //
    // &
    //
    position = pos(TEXT("&"), s);
    while (position != -1) {        
        ins(TEXT("&amp;"), s, position);
        del(s, position + 5, 1);        
        position = position + 2;
        previousPosition = position;                
        position = pos(TEXT("&"), &s[position]);
        
        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }  

    //
    // <
    //    
    position = 0;
    previousPosition = 0;
    position = pos(TEXT("<"), s);
    while (position != -1) {        
        ins(TEXT("&lt;"), s, position);
        del(s, position + 4, 1);        
        position = position + 2;
        previousPosition = position;                
        position = pos(TEXT("<"), &s[position]);
        
        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }  

    //
    // >
    //    
    position = 0;
    previousPosition = 0;
    position = pos(TEXT(">"), s);
    while (position != -1) {        
        ins(TEXT("&gt;"), s, position);
        del(s, position + 4, 1);        
        position = position + 2;
        previousPosition = position;                
        position = pos(TEXT(">"), &s[position]);
        
        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }  
}





/*
* return a pointer to new. free memory, please...
*/
wchar_t* getElementContent(wchar_t* xml, wchar_t* tag, unsigned int* pos) {
    wchar_t* p1       = NULL;
    wchar_t* p2       = NULL;
    wchar_t* ret      = NULL;

    wchar_t openTag[50];
    wchar_t closeTag[50];

    unsigned int l = wcslen(tag);

    if (pos != NULL) {
        *pos = 0;
    }

    wsprintf(openTag, TEXT("<%s>"), tag);
    wsprintf(closeTag, TEXT("</%s>"), tag);

    p1 = wcsstr(xml, openTag);

    if (p1 == NULL) {
        //
        // This is abcxyz
        //
        goto finally;
    }
    p1 = p1+l+2;

    if (*p1 == 0) {
        //
        // This is abc<tag>\0
        //
        goto finally;
    }

    p2 = wcsstr(p1, closeTag);

    if (p2 == NULL) {
        //
        // This is abc<tag>xyz\0
        //
        goto finally;
    }

    ret = new wchar_t[p2-p1+1];

    wcsncpy(ret, p1, p2-p1);
    ret[p2-p1] = 0;

    if (pos != NULL) {
        *pos = p2-xml+l+3;
    }

    finally:

    return ret;

}





/*
* Function that convert wchar_t* to char*. It creates buffer with new operator
* and has to be descarded by the caller
* @param wtoken - the buffer that refers to wchar_t pointer
*/
char* wchar2char(wchar_t* wtoken) {
    
    return toMultibyte(wtoken);
}

/*
* Function that convert char* to wchar_t*. It creates buffer with new operator
* and has to be descarded by the caller
* @param token - the buffer that refers to char pointer
*/

wchar_t* char2wchar(char *token) {

	return toWideChar(token);
}


void writeTheFile(wchar_t* path, wchar_t* content) {

    FILE* f = NULL;
    char* p = NULL;

    p = wchar2char(path);

    f = fopen(p, "wb");

    if (f == NULL) {
       return;
    }
    fwrite(content, sizeof(wchar_t), wcslen(content), f);

    fclose(f);

}


wchar_t* readContentFromFile(wchar_t* path) {

	wchar_t wfilename [256];

    int position = 0;
    int len = 0;
	wsprintf(wfilename, TEXT("%s"),  path);
	HANDLE file;
    DWORD lpFileSizeHigh;
    DWORD dwSize;

	wchar_t line[2048];
	FILE* f;
	wchar_t* ptr = NULL;
    wchar_t* ret = NULL;

    if (dbg) {
        MessageBox (NULL, path, TEXT ("readContentFromFile"), MB_SETFOREGROUND |MB_OK);
    }

    file = CreateFile(wfilename, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
                              NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);


    if( file ) {
        dwSize = GetFileSize(file, &lpFileSizeHigh);
        ptr = new wchar_t[dwSize + 1];
        wmemset(ptr, 0, dwSize);
    } else {

        goto finally;
    }

    CloseHandle(file);

    //f = _wfopen(path, TEXT("r"));
    f = _wfopen(path, TEXT("rb"));
    if (f == NULL) {
        goto finally;
    }

    while(fgetws(line, 2048, f) != NULL) {

        len = wcslen(line);
        wcsncpy(&ptr[position], line, len);

        position = position + len;

    }

    fflush(f);
    fclose(f);

finally:

    if (ptr) {
        ret = _wcsdup(&ptr[1]);         // To remove 1st newline character (see fgetws specs)
        len = wcslen(ret);
        ret[len-2] = 0;                 // To remove last 2 newline characters (appended by VB)

        delete [] ptr; ptr = NULL;
    }
    return ret;
    // return ptr;

}


/*
* function to replace &lt; or &gt; to < or > from server response in syncML message. New function for SyncServer 2.0
*/
void replaceAmpLGt(string &s) {

    if (dbg) {
            MessageBox (NULL, s, TEXT("in replaceAmpLGt before"), MB_SETFOREGROUND |MB_OK);
    }

    int position = 0;

    // search for &lt;
    position = pos(TEXT("&lt;"), s);
    while (position != -1) {
        del(s, position, 4);
        ins(TEXT("<"), s, position);
        position = pos(TEXT("&lt;"), s);
    }
    // now search for &lt;
    position = 0;
    position = pos(TEXT("&gt;"), s);
    while (position != -1) {
        del(s, position, 4);
        ins(TEXT(">"), s, position);
        position = pos(TEXT("&gt;"), s);
    }
    // now search for &amp;
    position = 0;
    position = pos(TEXT("&amp;"), s);
    while (position != -1) {
        del(s, position, 5);
        ins(TEXT("&"), s, position);
        position = pos(TEXT("&amp;"), s);
    }
    if (dbg) {
        MessageBox (NULL, s, TEXT("in replaceAmpLGt after"), MB_SETFOREGROUND |MB_OK);
    }

}

/*
* Function to modify "&" with "&amp;" , "<" with "lt;" and ">" with "gt;"
* to prevent server crash
*/

void encodeSpecialChar2XML(string &s) {

    int position = 0, previousPosition = 0;

    //
    // &
    //
    position = pos(TEXT("&"), s);
    while (position != -1) {
        del(s, position, 1);
        ins(TEXT("&amp;"), s, position);
        position = position + 2;
        previousPosition = position;
        position = pos(TEXT("&"), &s[position]);

        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }

    //
    // <
    //
    position = 0;
    previousPosition = 0;
    position = pos(TEXT("<"), s);
    while (position != -1) {
        del(s, position, 1);
        ins(TEXT("&lt;"), s, position);
        position = position + 2;
        previousPosition = position;
        position = pos(TEXT("<"), &s[position]);

        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }

    //
    // >
    //
    position = 0;
    previousPosition = 0;
    position = pos(TEXT(">"), s);
    while (position != -1) {
        del(s, position, 1);
        ins(TEXT("&gt;"), s, position);
        position = position + 2;
        previousPosition = position;
        position = pos(TEXT(">"), &s[position]);

        if (position == -1) {
            break;
        }
        position = previousPosition + position;
    }

}


wchar_t* encodeBase64UTF8(char* toEncode) {

    char*    base64   = NULL;
    int      len      = 0;
    wchar_t* toReturn = NULL;
    wchar_t* tmp      = NULL;

    len = strlen(toEncode);
    base64 = new char[2 * len + 1];

    b64_encode(base64, toEncode, len);

    toReturn = char2wchar(base64);

    if (base64 != NULL)
        delete [] base64;

    return toReturn;

}


/*
* Return the encoded wchar_t array.
*
* Function uses to encode wchar_t array into base64.
*
* @param toEncode - the wchar_t array to encode
*/

wchar_t* encodeBase64W(wchar_t* toEncode) {

    char*    c        = NULL;
    char*    base64   = NULL;
    int      len      = 0;
    wchar_t* toReturn = NULL;
    wchar_t* tmp      = NULL;

    c = wchar2char(toEncode);
    if (c == NULL) {
        return NULL;
    }

    len = strlen(c);
    base64 = new char[2 * len + 1];

    b64_encode(base64, c, len);

    toReturn = char2wchar(base64);

    if (c != NULL)
        delete [] c;

    if (base64 != NULL)
        delete [] base64;

    return toReturn;

}

wchar_t* decodeBase64WUTF8(wchar_t* toDecode) {

    int      num        = 0;
    byte*    reversed   = NULL;
    char*    content    = NULL;
    wchar_t* toReturn   = NULL;
    int k = 0;

    content = wchar2char(toDecode);

    if (content == NULL) {
        return NULL;
    }

    reversed = new byte[2 * strlen(content) + 1];
    memset(reversed, 0, 2 * strlen(content) + 1);

    num = strlen(content);
    b64_decode(reversed, content);
    reversed[num] = 0;

    k = MultiByteToWideChar(CP_UTF8, 0, (char*)reversed, -1, NULL, 0);

    //
    // MultiByteToWideChar returns the number of byte required including the terminator too
    //
    toReturn = new wchar_t[k];
    wmemset(toReturn, 0, k);

    MultiByteToWideChar(CP_UTF8, 0, (char*)reversed, num, toReturn, k);

    if (content != NULL)
        delete [] content;

    if (reversed != NULL)
        delete [] reversed;

    return toReturn;

}


/*
* Return the decoded wchar_t array.
*
* Function uses to dencode wchar_t array from base64.
*
* @param toDecode - the wchar_t array to decode
*/

wchar_t* decodeBase64W(wchar_t* toDecode) {

    int      num        = 0;
    byte*    reversed   = NULL;
    char*    content    = NULL;
    wchar_t* toReturn   = NULL;

    content = wchar2char(toDecode);

    if (content == NULL) {
        return NULL;
    }

    reversed = new byte[2 * strlen(content) + 1];
    memset(reversed, 0, 2 * strlen(content) + 1);

    num = strlen(content);
    b64_decode(reversed, content);
    reversed[num] = 0;

    toReturn = char2wchar((char*)reversed);

    if (content != NULL)
        delete [] content;

    if (reversed != NULL)
        delete [] reversed;

    return toReturn;

}

//void error() {
//    //wsprintf(logmsg, TEXT("Error!\nerror code: %d\nerror message: %s"), lastErrorCode, lastErrorMsg);
//    //LOG.error(logmsg);
//}
//// to return messages to user...
//void error(string &errorMessage, int code, wchar_t* errMsg) {
//    wchar_t localTemp[DIM_LOCAL_TEMP];
//    wsprintf(localTemp, TEXT("Error!\nerror code: %d\nerror message: %s"), code, errMsg);
//
//    errorMessage = errorMessage + localTemp;
//    ActionLOG.action((wchar_t*)errorMessage);
//}
//void error(string &errorMessage) {
//    error(errorMessage, lastErrorCode,lastErrorMsg);
//}



/*
 * Check if we are using SIF format for the source 'sourceName'.
 */
bool checkSIF(const wchar_t* wsourceName) {

    int ret;
    OutlookConfig oc;
    bool useSIF = true;
	char* internalBuffer = new char[MAX_KEY_LENGTH];
	char* syncSourceName = new char[DIM_LOCAL_TEMP];

    // Set the appropriate syncSourceName
    if (wcscmp(wsourceName, CONTACTS) == 0){			  // SYNC_CONTACTS  
        strcpy(syncSourceName, SYNC_CONTACTS);	
    }                                                           
    else if (wcscmp(wsourceName, APPOINTMENTS) == 0){     // SYNC_APPOINTMENTS
        strcpy(syncSourceName, SYNC_APPOINTMENTS);
    }                                                     
    else if (wcscmp(wsourceName, TASKS) == 0){            // SYNC_TASKS
		strcpy(syncSourceName, SYNC_TASKS);
    }                                                    
    else if (wcscmp(wsourceName, NOTES) == 0){            // SYNC_NOTES
		strcpy(syncSourceName, SYNC_NOTES);
    }
    else{
		LOG.error("CheckSIF: invalid source name: %ls", wsourceName);
        goto finally;
    }

    oc = OutlookConfig();
    ret = oc.getSourcePropertyValue(syncSourceName, PROPERTY_USESIF, internalBuffer);


	// Registry key not found / errors -> default = SIF
	if (ret < 0) {
		useSIF = true;
		goto finally;
	}

	if(strcmp(internalBuffer, T("0")) == 0)
		useSIF = false;
	else
		useSIF = true;		// Other value -> default = SIF

finally:

	delete [] internalBuffer;
	delete [] syncSourceName;

    return useSIF;
}
